#include "AStar.h"

#include <algorithm>
#include <optional>

#include <VirtualRobot/Visualization/CoinVisualization/CoinVisualizationFactory.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/aron/core/navigator/data/container/Dict.h>
#include <RobotAPI/libraries/aron/core/navigator/data/primitive/String.h>

#include <armarx/navigation/algorithms/astar/AStarPlanner.h>
#include <armarx/navigation/algorithms/smoothing/ChainApproximation.h>
#include <armarx/navigation/algorithms/smoothing/CircularPathSmoothing.h>
#include <armarx/navigation/conversions/eigen.h>
#include <armarx/navigation/core/Trajectory.h>
#include <armarx/navigation/core/constants.h>
#include <armarx/navigation/global_planning/GlobalPlanner.h>
#include <armarx/navigation/global_planning/aron/AStarParams.aron.generated.h>
#include <armarx/navigation/global_planning/aron_conversions.h>
#include <armarx/navigation/global_planning/core.h>
#include <armarx/navigation/global_planning/optimization/OrientationOptimizer.h>


namespace armarx::navigation::glob_plan
{

    // AStarParams

    Algorithms
    AStarParams::algorithm() const
    {
        return Algorithms::AStar;
    }

    aron::datanavigator::DictNavigatorPtr
    AStarParams::toAron() const
    {
        arondto::AStarParams dto;

        AStarParams bo;
        aron_conv::toAron(dto, bo);

        return dto.toAron();
    }

    AStarParams
    AStarParams::FromAron(const aron::datanavigator::DictNavigatorPtr& dict)
    {
        ARMARX_CHECK_NOT_NULL(dict);

        ARMARX_INFO << dict->getAllKeysAsString();

        arondto::AStarParams dto;
        dto.fromAron(dict);

        AStarParams bo;
        aron_conv::fromAron(dto, bo);

        return bo;
    }

    // AStar

    AStar::AStar(const AStarParams& params, const core::Scene& ctx) :
        GlobalPlanner(ctx), params(params)
    {
    }

    auto
    robotCollisionModel()
    {
        VirtualRobot::CoinVisualizationFactory factory;
        const auto cylinder = factory.createCylinder(1500.F / 2, 2000.F);

        VirtualRobot::CollisionModelPtr collisionModel(new VirtualRobot::CollisionModel(cylinder));

        return collisionModel;
    }

    std::vector<Eigen::Vector2f>
    AStar::postProcessPath(const std::vector<Eigen::Vector2f>& path)
    {
        /// chain approximation
        algorithm::ChainApproximation approx(path,
                                        algorithm::ChainApproximation::Params{.distanceTh = 200.F});
        approx.approximate();
        const auto p = approx.approximatedChain();

        // visualizePath(p, "approximated", simox::Color::green());

        // algo::CircularPathSmoothing smoothing;
        // const auto points = smoothing.smooth(p);

        return p;
    }

    std::optional<GlobalPlannerResult>
    AStar::plan(const core::Pose& goal)
    {
        ARMARX_TRACE;

        algorithm::astar::AStarPlanner planner(scene.robot, scene.staticScene->objects, 100.F);
        // planner.setRobotColModel("Platform-colmodel");

        // planner.setRobotColModel("Platform-colmodel");
        planner.setRobotCollisionModel(robotCollisionModel());

        const Eigen::Vector2f startPos = conv::to2D(scene.robot->getGlobalPosition());
        const Eigen::Vector2f goalPos = conv::to2D(goal.translation());

        const auto timeStart = IceUtil::Time::now();

        std::vector<Eigen::Vector2f> plan;
        try
        {
            plan = planner.plan(startPos, goalPos);
        }
        catch (...)
        {
            ARMARX_WARNING << "Could not plan collision-free path.";
            return std::nullopt;
        }
        const auto timeEnd = IceUtil::Time::now();

        ARMARX_IMPORTANT << "A* planning took " << (timeEnd - timeStart).toMilliSeconds() << " ms";
        ARMARX_IMPORTANT << "Path contains " << plan.size() << " points";

        if (plan.size() < 2) // failure
        {
            ARMARX_WARNING << "Could not plan collision-free path!";
            return std::nullopt;
        }

        ARMARX_TRACE;
        auto smoothPlan = postProcessPath(plan);
        ARMARX_IMPORTANT << "Smooth path contains " << smoothPlan.size() << " points";

        ARMARX_TRACE;
        // we need to strip the first and the last points from the plan as they encode the start and goal position
        smoothPlan.erase(smoothPlan.begin());
        smoothPlan.pop_back();

        ARMARX_TRACE;
        auto trajectory = core::Trajectory::FromPath(core::Pose(scene.robot->getGlobalPose()),
                                                     conv::to3D(smoothPlan),
                                                     goal,
                                                     params.linearVelocity);

        // TODO(fabian.reister): resampling of trajectory

        // FIXME param
        auto resampledTrajectory = trajectory.resample(params.resampleDistance);
        ARMARX_IMPORTANT << "Resampled trajectory contains " << resampledTrajectory.points().size()
                         << " points";

        // ARMARX_CHECK(resampledTrajectory.hasMaxDistanceBetweenWaypoints(params.resampleDistance));


        ARMARX_TRACE;
        OrientationOptimizer optimizer(resampledTrajectory, OrientationOptimizer::Params{});
        const auto result = optimizer.optimize();

        if (not result)
        {
            ARMARX_ERROR << "Optimizer failure";
            return std::nullopt;
        }

        // TODO circular path smoothing should be done now

        algorithm::CircularPathSmoothing smoothing;
        const auto smoothTrajectory = smoothing.smooth(result.trajectory);

        ARMARX_TRACE;
        return GlobalPlannerResult{.trajectory = smoothTrajectory};
    }
} // namespace armarx::navigation::glob_plan
