/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Graph.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/wm/memory_definitions.h>

#include <armarx/navigation/location/aron/Location.aron.generated.h>


namespace armarx::navigation::graph
{

    std::string
    VertexAttribs::getName() const
    {
        return aron.locationID.providerSegmentName + "/" + aron.locationID.entityName;
    }

    bool
    VertexAttribs::hasPose() const
    {
        return _pose.has_value();
    }

    core::Pose
    VertexAttribs::getPose() const
    {
        ARMARX_CHECK(_pose.has_value());
        return _pose.value();
    }

    void
    VertexAttribs::setPose(const core::Pose& pose)
    {
        this->_pose = pose;
    }

} // namespace armarx::navigation::graph


namespace armarx::navigation
{

    void
    graph::toAron(arondto::Graph& dto, const Graph& bo)
    {
        dto = {};
        for (auto vertex : bo.vertices())
        {
            auto& v = dto.vertices.emplace_back(vertex.attrib().aron);
            v.vertexID = static_cast<long>(vertex.objectID());
        }
        ARMARX_CHECK_EQUAL(dto.vertices.size(), bo.numVertices());

        for (auto edge : bo.edges())
        {
            auto& e = dto.edges.emplace_back(edge.attrib().aron);
            e.sourceVertexID = static_cast<long>(edge.sourceObjectID());
            e.targetVertexID = static_cast<long>(edge.targetObjectID());
        }
        ARMARX_CHECK_EQUAL(dto.edges.size(), bo.numEdges());
    }


    void
    graph::fromAron(const arondto::Graph& dto, Graph& bo)
    {
        bo = {};
        for (const arondto::Vertex& vertex : dto.vertices)
        {
            auto v = bo.addVertex(semrel::ShapeID(vertex.vertexID));
            v.attrib().aron = vertex;
        }
        ARMARX_CHECK_EQUAL(bo.numVertices(), dto.vertices.size());

        for (const arondto::Edge& edge : dto.edges)
        {
            auto e = bo.addEdge(semrel::ShapeID(edge.sourceVertexID),
                                semrel::ShapeID(edge.targetVertexID));
            e.attrib().aron = edge;
        }
        ARMARX_CHECK_EQUAL(bo.numEdges(), dto.edges.size());
    }


    void
    graph::resolveLocation(Graph::Vertex& vertex,
                           const aron::datanavigator::DictNavigatorPtr& locationData)
    {
        navigation::location::arondto::Location dto;
        dto.fromAron(locationData);
        vertex.attrib().setPose(core::Pose(dto.globalRobotPose));
    }


} // namespace armarx::navigation
