#include "Navigator.h"

#include <algorithm>

#include <Eigen/Geometry>

#include <SimoxUtility/math/convert/mat3f_to_rpy.h>
#include <VirtualRobot/Robot.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/util/CPPUtility/trace.h>

#include <armarx/navigation/core/events.h>
#include <armarx/navigation/core/time/TimeServerInterface.h>
#include <armarx/navigation/core/types.h>
#include <armarx/navigation/server/StackResult.h>
#include <armarx/navigation/server/monitoring/GoalReachedMonitor.h>

namespace armarx::navigation::server
{
    void
    stopIfRunning(PeriodicTask<Navigator>::pointer_type& task)
    {
        if (task)
        {
            task->stop();
        }
    }

    Navigator::Navigator(const Config& config, const InjectedServices& services) :
        config{config}, srv{services}
    {
        ARMARX_CHECK_NOT_NULL(config.scene) << "The scene must be set!";
        ARMARX_CHECK_NOT_NULL(services.executor) << "The executor service must be set!";
        ARMARX_CHECK_NOT_NULL(services.publisher) << "The publisher service must be set!";
    }

    Navigator::~Navigator()
    {
        ARMARX_INFO << "Navigator destructor";
        stop();

        stopAllThreads();
    }

    void
    Navigator::moveTo(const std::vector<core::Pose>& waypoints,
                      core::NavigationFrame navigationFrame)
    {
        ARMARX_INFO << "Received moveTo() request.";

        std::vector<core::Pose> globalWaypoints;
        switch (navigationFrame)
        {
            case core::NavigationFrame::Absolute:
                globalWaypoints = waypoints;
                break;
            case core::NavigationFrame::Relative:
                globalWaypoints.reserve(waypoints.size());
                std::transform(std::begin(waypoints),
                               std::end(waypoints),
                               std::back_inserter(globalWaypoints),
                               [&](const core::Pose& p)
                               { return core::Pose(config.scene->robot->getGlobalPose()) * p; });
                break;
        }

        moveToAbsolute(globalWaypoints);
    }

    void
    Navigator::moveToAbsolute(const std::vector<core::Pose>& waypoints)
    {
        ARMARX_TRACE;
        // if this navigator is in use, stop the movement ...
        pause();
        // ... and all threads
        stopAllThreads();

        ARMARX_TRACE;

        // global planner
        ARMARX_INFO << "Planning global trajectory";
        ARMARX_CHECK_NOT_NULL(config.stack.globalPlanner);
        // TODO plan on multiple waypoints, ignoring waypoints for now
        // idea: compute multiple global trajectories, one for each segment between waypoints.
        globalPlan = config.stack.globalPlanner->plan(waypoints.back());

        ARMARX_TRACE;

        if (not globalPlan.has_value())
        {
            ARMARX_WARNING << "No global trajectory. Cannot move.";
            srv.publisher->globalPlanningFailed(core::GlobalPlanningFailedEvent{
                {.timestamp = config.scene->timeServer->now()}, {""}});
            return;
        }
        ARMARX_TRACE;

        srv.publisher->globalTrajectoryUpdated(globalPlan.value());

        ARMARX_INFO << "Global planning completed. Will now start all required threads";
        ARMARX_TRACE;

        // local planner
        if (config.stack.localPlanner)
        {
            ARMARX_INFO << "Planning local trajectory enabled";
            localPlannerTask =
                new PeriodicTask<Navigator>(this,
                                            &Navigator::updateLocalPlanner,
                                            config.general.tasks.localPlanningUpdatePeriod,
                                            false,
                                            "LocalPlannerTask");
            localPlannerTask->start();
        }

        // trajectory controller
        ARMARX_CHECK_NOT_NULL(config.stack.trajectoryControl);
        trajectoryControllerTask =
            new PeriodicTask<Navigator>(this,
                                        &Navigator::updateTrajectoryController,
                                        config.general.tasks.trajectoryControllerUpdatePeriod,
                                        false,
                                        "TrajectoryControllerTask");
        trajectoryControllerTask->start();

        // safety controller
        if (config.stack.safetyControl)
        {
            ARMARX_INFO << "Safety control enabled";
            safetyControllerTask =
                new PeriodicTask<Navigator>(this,
                                            &Navigator::updateSafetyController,
                                            config.general.tasks.safetyControllerUpdatePeriod,
                                            false,
                                            "SafetyControllerTask");
            safetyControllerTask->start();
        }

        // executor
        executorTask = new PeriodicTask<Navigator>(this,
                                                   &Navigator::updateExecutor,
                                                   config.general.tasks.executorUpdatePeriod,
                                                   false,
                                                   "ExecutorTask");
        executorTask->start();

        // introspector
        introspectorTask =
            new PeriodicTask<Navigator>(this,
                                        &Navigator::updateIntrospector,
                                        config.general.tasks.introspectorUpdatePeriod,
                                        false,
                                        "IntrospectorTask");
        introspectorTask->start();

        // monitoring
        ARMARX_INFO << "Starting monitoring";
        goalReachedMonitor =
            GoalReachedMonitor(waypoints, *config.scene, GoalReachedMonitorConfig());
        monitorTask = new PeriodicTask<Navigator>(this,
                                                  &Navigator::updateMonitor,
                                                  config.general.tasks.monitorUpdatePeriod,
                                                  false,
                                                  "MonitorTask");
        monitorTask->start();

        // Could be required if pauseMovement() has been called in the past.
        resume();
        srv.publisher->movementStarted(core::MovementStartedEvent{
            {.timestamp = config.scene->timeServer->now()}, {.startPose = waypoints.back()}});
    }

    void
    Navigator::moveTowards(const core::Direction& direction, core::NavigationFrame navigationFrame)
    {
    }

    void
    Navigator::moveTowardsAbsolute(const core::Direction& direction)
    {
    }

    void
    Navigator::updateLocalPlanner()
    {
        localPlan = config.stack.localPlanner->plan(globalPlan->trajectory);

        if (localPlan.has_value())
        {
            srv.publisher->localTrajectoryUpdated(localPlan.value());
        }
        else
        {
            // srv.publisher->localTrajectoryPlanningFailed();
        }
    }

    void
    Navigator::updateTrajectoryController()
    {
        // TODO distinguish between global and local plan!

        if (not globalPlan.has_value())
        {
            return;
        }

        trajectoryCtrlResult = config.stack.trajectoryControl->control(globalPlan->trajectory);

        srv.publisher->trajectoryControllerUpdated(trajectoryCtrlResult.value());
    }

    void
    Navigator::updateSafetyController()
    {
        // FIXME
        // if (res.controlVelocity)
        // {
        //     res.safeVelocity = config.stack.safetyControl->control(res.controlVelocity.value());
        // }
    }

    void
    Navigator::updateExecutor()
    {
        if (isPaused() or isStopped())
        {
            // [[unlikely]]
            ARMARX_INFO << deactivateSpam(1) << "stopped or paused";
            return;
        }

        if (not isStackResultValid())
        {
            ARMARX_INFO << deactivateSpam(1) << "stack result invalid";
            // [[unlikely]]
            return;
        }

        const core::Twist twist = [&]() -> core::Twist
        {
            if (config.stack.safetyControl != nullptr)
            {
                return safetyCtrlResult->twist;
            }

            return trajectoryCtrlResult->twist;
        }();

        const core::Rotation robot_R_world(config.scene->robot->getGlobalOrientation().inverse());

        ARMARX_INFO << deactivateSpam(100) << "Robot orientation "
                    << simox::math::mat3f_to_rpy(config.scene->robot->getGlobalOrientation()).z();

        core::Twist robotFrameVelocity;

        robotFrameVelocity.linear = robot_R_world * twist.linear;
        // FIXME fix angular velocity
        robotFrameVelocity.angular = twist.angular;

        ARMARX_INFO << deactivateSpam(1) << "velocity in robot frame " << robotFrameVelocity.linear;

        srv.executor->move(robotFrameVelocity);
    }

    void
    Navigator::updateIntrospector()
    {
        ARMARX_CHECK_NOT_NULL(srv.introspector);

        // if (res.isValid())
        // {
        //     srv.introspector->onStackResult(res);
        // }
    }

    void
    Navigator::updateMonitor()
    {
        if (not isPaused() and goalReachedMonitor->goalReached())
        {
            // [[unlikely]]

            ARMARX_INFO << "Goal reached!";
            stop();

            // stop all threads, including this one
            stopAllThreads();
        }
    }

    void
    Navigator::stopAllThreads()
    {
        stopIfRunning(localPlannerTask);
        stopIfRunning(trajectoryControllerTask);
        stopIfRunning(safetyControllerTask);
        stopIfRunning(executorTask);
        stopIfRunning(introspectorTask);
        stopIfRunning(monitorTask);
    }

    bool
    Navigator::isStackResultValid() const noexcept
    {

        // global planner
        if (config.stack.globalPlanner != nullptr)
        {
            if (not globalPlan.has_value())
            {
                ARMARX_WARNING << deactivateSpam(1) << "Global trajectory not yet set.";
                return false;
            }
        }

        // local planner
        if (config.stack.localPlanner != nullptr)
        {
            if (not localPlan.has_value())
            {
                ARMARX_WARNING << deactivateSpam(1) << "Local trajectory not yet set.";
                return false;
            }
        }

        // trajectory controller
        if (config.stack.trajectoryControl != nullptr)
        {
            if (not trajectoryCtrlResult.has_value())
            {
                ARMARX_WARNING << deactivateSpam(1) << "Raw control velocity not yet set.";
                return false;
            }
        }

        // safety controller
        if (config.stack.safetyControl != nullptr)
        {
            if (not safetyCtrlResult.has_value())
            {
                ARMARX_WARNING << deactivateSpam(1) << "Safe velocity not yet set.";
                return false;
            }
        }

        // [[likely]]
        return true;
    }

    void
    Navigator::pause()
    {
        ARMARX_INFO << "pause";

        executorEnabled.store(false);

        ARMARX_CHECK_NOT_NULL(srv.executor);
        srv.executor->move(core::Twist::Zero());
    }

    void
    Navigator::resume()
    {
        ARMARX_INFO << "resume";

        executorEnabled.store(true);
    }

    void
    Navigator::stop()
    {
        ARMARX_INFO << "stop";

        pause();
        // TODO: Reset own state.
    }

    bool
    Navigator::isPaused() const noexcept
    {
        return not executorEnabled.load();
    }

    bool
    Navigator::isStopped() const noexcept
    {
        // TODO: Implement.
        return false;
    }

    Navigator::Navigator(Navigator&& other) noexcept :
        config{std::move(other.config)},
        srv{std::move(other.srv)},
        executorEnabled{other.executorEnabled.load()}
    {
    }
} // namespace armarx::navigation::server
